/**
 * Implementation of the net.sf.geographiclib.Geodesic class
 *
 * Copyright (c) Charles Karney (2013-2022) <charles></charles>@karney.com> and licensed
 * under the MIT/X11 License.  For more information, see
 * https://geographiclib.sourceforge.io/
 */
package net.sf.geographiclib

import net.sf.geographiclib.GeographicErr
import kotlin.jvm.JvmField
import kotlin.jvm.JvmOverloads

/**
 * Geodesic calculations.
 *
 *
 * The shortest path between two points on an ellipsoid at (*lat1*,
 * *lon1*) and (*lat2*, *lon2*) is called the geodesic.  Its
 * length is *s12* and the geodesic from point 1 to point 2 has azimuths
 * *azi1* and *azi2* at the two end points.  (The azimuth is the
 * heading measured clockwise from north.  *azi2* is the "forward"
 * azimuth, i.e., the heading that takes you beyond point 2 not back to point
 * 1.)
 *
 *
 * Given *lat1*, *lon1*, *azi1*, and *s12*, we can
 * determine *lat2*, *lon2*, and *azi2*.  This is the
 * *direct* geodesic problem and its solution is given by the function
 * [Direct][.Direct].  (If *s12* is sufficiently large that the
 * geodesic wraps more than halfway around the earth, there will be another
 * geodesic between the points with a smaller *s12*.)
 *
 *
 * Given *lat1*, *lon1*, *lat2*, and *lon2*, we can
 * determine *azi1*, *azi2*, and *s12*.  This is the
 * *inverse* geodesic problem, whose solution is given by [ Inverse][.Inverse].  Usually, the solution to the inverse problem is unique.  In cases
 * where there are multiple solutions (all with the same *s12*, of
 * course), all the solutions can be easily generated once a particular
 * solution is provided.
 *
 *
 * The standard way of specifying the direct problem is the specify the
 * distance *s12* to the second point.  However it is sometimes useful
 * instead to specify the arc length *a12* (in degrees) on the auxiliary
 * sphere.  This is a mathematical construct used in solving the geodesic
 * problems.  The solution of the direct problem in this form is provided by
 * [ArcDirect][.ArcDirect].  An arc length in excess of 180 indicates
 * that the geodesic is not a shortest path.  In addition, the arc length
 * between an equatorial crossing and the next extremum of latitude for a
 * geodesic is 90.
 *
 *
 * This class can also calculate several other quantities related to
 * geodesics.  These are:
 *
 *  *
 * *reduced length*.  If we fix the first point and increase
 * *azi1* by *dazi1* (radians), the second point is displaced
 * *m12* *dazi1* in the direction *azi2* + 90.  The
 * quantity *m12* is called the "reduced length" and is symmetric under
 * interchange of the two points.  On a curved surface the reduced length
 * obeys a symmetry relation, *m12* + *m21* = 0.  On a flat
 * surface, we have *m12* = *s12*.  The ratio *s12* / *m12*
 * gives the azimuthal scale for an azimuthal equidistant projection.
 *  *
 * *geodesic scale*.  Consider a reference geodesic and a second
 * geodesic parallel to this one at point 1 and separated by a small distance
 * *dt*.  The separation of the two geodesics at point 2 is *M12*
 * *dt* where *M12* is called the "geodesic scale".  *M21* is
 * defined similarly (with the geodesics being parallel at point 2).  On a
 * flat surface, we have *M12* = *M21* = 1.  The quantity
 * 1/ *M12* gives the scale of the Cassini-Soldner projection.
 *  *
 * *area*.  The area between the geodesic from point 1 to point 2 and
 * the equation is represented by *S12*; it is the area, measured
 * counter-clockwise, of the geodesic quadrilateral with corners
 * (*lat1*,*lon1*), (0,*lon1*), (0,*lon2*), and
 * (*lat2*,*lon2*).  It can be used to compute the area of any
 * geodesic polygon.
 *
 *
 *
 * The quantities *m12*, *M12*, *M21* which all specify the
 * behavior of nearby geodesics obey addition rules.  If points 1, 2, and 3 all
 * lie on a single geodesic, then the following rules hold:
 *
 *  *
 * *s13* = *s12* + *s23*
 *  *
 * *a13* = *a12* + *a23*
 *  *
 * *S13* = *S12* + *S23*
 *  *
 * *m13* = *m12* *M23* + *m23* *M21*
 *  *
 * *M13* = *M12* *M23*  (1  *M12*
 * *M21*) *m23* / *m12*
 *  *
 * *M31* = *M32* *M21*  (1  *M23*
 * *M32*) *m12* / *m23*
 *
 *
 *
 * The results of the geodesic calculations are bundled up into a [ ] object which includes the input parameters and all the
 * computed results, i.e., *lat1*, *lon1*, *azi1*, *lat2*,
 * *lon2*, *azi2*, *s12*, *a12*, *m12*, *M12*,
 * *M21*, *S12*.
 *
 *
 * The functions [Direct][.Direct],
 * [ArcDirect][.ArcDirect], and
 * [Inverse][.Inverse] include an
 * optional final argument *outmask* which allows you specify which
 * results should be computed and returned.  If you omit *outmask*, then
 * the "standard" geodesic results are computed (latitudes, longitudes,
 * azimuths, and distance).  *outmask* is bitor'ed combination of [ ] values.  For example, if you wish just to compute the distance
 * between two points you would call, e.g.,
 * <pre>
 * `GeodesicData g = Geodesic.WGS84.Inverse(lat1, lon1, lat2, lon2,
 * GeodesicMask.DISTANCE); `</pre>
 *
 *
 * Additional functionality is provided by the [GeodesicLine] class,
 * which allows a sequence of points along a geodesic to be computed.
 *
 *
 * The shortest distance returned by the solution of the inverse problem is
 * (obviously) uniquely defined.  However, in a few special cases there are
 * multiple azimuths which yield the same shortest distance.  Here is a
 * catalog of those cases:
 *
 *  *
 * *lat1* = *lat2* (with neither point at a pole).  If
 * *azi1* = *azi2*, the geodesic is unique.  Otherwise there are
 * two geodesics and the second one is obtained by setting [*azi1*,
 * *azi2*]  [*azi2*, *azi1*], [*M12*, *M21*]
 *  [*M21*, *M12*], *S12*  *S12*.
 * (This occurs when the longitude difference is near 180 for
 * oblate ellipsoids.)
 *  *
 * *lon2* = *lon1*  180 (with neither point at a
 * pole).  If *azi1* = 0 or 180, the geodesic is
 * unique.  Otherwise there are two geodesics and the second one is obtained
 * by setting [ *azi1*, *azi2*]  [*azi1*,
 * *azi2*], *S12*   *S12*.  (This occurs
 * when *lat2* is near *lat1* for prolate ellipsoids.)
 *  *
 * Points 1 and 2 at opposite poles.  There are infinitely many geodesics
 * which can be generated by setting [*azi1*, *azi2*]
 * [*azi1*, *azi2*] + [*d*, *d*], for arbitrary
 * *d*.  (For spheres, this prescription applies when points 1 and 2 are
 * antipodal.)
 *  *
 * *s12* = 0 (coincident points).  There are infinitely many geodesics
 * which can be generated by setting [*azi1*, *azi2*]
 * [*azi1*, *azi2*] + [*d*, *d*], for arbitrary *d*.
 *
 *
 *
 * The calculations are accurate to better than 15 nm (15 nanometers) for the
 * WGS84 ellipsoid.  See Sec. 9 of
 * [arXiv:1102.1215v1](https://arxiv.org/abs/1102.1215v1) for
 * details.  The algorithms used by this class are based on series expansions
 * using the flattening *f* as a small parameter.  These are only accurate
 * for |*f*| &lt; 0.02; however reasonably accurate results will be
 * obtained for |*f*| &lt; 0.2.  Here is a table of the approximate
 * maximum error (expressed as a distance) for an ellipsoid with the same
 * equatorial radius as the WGS84 ellipsoid and different values of the
 * flattening.<pre>
 * |f|      error
 * 0.01     25 nm
 * 0.02     30 nm
 * 0.05     10 um
 * 0.1     1.5 mm
 * 0.2     300 mm </pre>
 *
 *
 * The algorithms are described in
 *
 *  * C. F. F. Karney,
 * [
 * Algorithms for geodesics](https://doi.org/10.1007/s00190-012-0578-z),
 * J. Geodesy **87**, 4355 (2013)
 * ([
 * addenda](https://geographiclib.sourceforge.io/geod-addenda.html)).
 *
 *
 *
 * Example of use:
 * <pre>
 * `// Solve the direct geodesic problem.
 *
 * // This program reads in lines with lat1, lon1, azi1, s12 and prints
 * // out lines with lat2, lon2, azi2 (for the WGS84 ellipsoid).
 *
 * import java.util.*;
 * import net.sf.geographiclib.*;
 * public class Direct {
 * public static void main(String[] args) {
 * try {
 * Scanner in = new Scanner(System.in);
 * double lat1, lon1, azi1, s12;
 * while (true) {
 * lat1 = in.nextDouble(); lon1 = in.nextDouble();
 * azi1 = in.nextDouble(); s12 = in.nextDouble();
 * GeodesicData g = Geodesic.WGS84.Direct(lat1, lon1, azi1, s12);
 * System.out.println(g.lat2 + " " + g.lon2 + " " + g.azi2);
 * }
 * }
 * catch (Exception e) {}
 * }
 * }`</pre>
 */
class Geodesic(var _a: Double, var _f: Double) {
    var _f1: Double
    var _e2: Double
    var _ep2: Double
    var _b: Double
    var _c2: Double
    private val _n: Double
    private val _etol2: Double
    private val _A3x: DoubleArray
    private val _C3x: DoubleArray
    private val _C4x: DoubleArray

    /**
     * Solve the direct geodesic problem where the length of the geodesic
     * is specified in terms of distance.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param s12 distance between point 1 and point 2 (meters); it can be
     * negative.
     * @return a [GeodesicData] object with the following fields:
     * *lat1*, *lon1*, *azi1*, *lat2*, *lon2*,
     * *azi2*, *s12*, *a12*.
     *
     *
     * *lat1* should be in the range [90, 90].  The values
     * of *lon2* and *azi2* returned are in the range [180,
     * 180].
     *
     *
     * If either point is at a pole, the azimuth is defined by keeping the
     * longitude fixed, writing *lat* = (90  ),
     * and taking the limit   0+.  An arc length greater that
     * 180 signifies a geodesic which is not a shortest path.  (For a
     * prolate ellipsoid, an additional condition is necessary for a shortest
     * path: the longitudinal extent must not exceed of 180.)
     */
    fun Direct(
        lat1: Double, lon1: Double,
        azi1: Double, s12: Double
    ): GeodesicData {
        return Direct(lat1, lon1, azi1, false, s12, GeodesicMask.STANDARD)
    }

    /**
     * Solve the direct geodesic problem where the length of the geodesic is
     * specified in terms of distance and with a subset of the geodesic results
     * returned.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param s12 distance between point 1 and point 2 (meters); it can be
     * negative.
     * @param outmask a bitor'ed combination of [GeodesicMask] values
     * specifying which results should be returned.
     * @return a [GeodesicData] object with the fields specified by
     * *outmask* computed.
     *
     *
     * *lat1*, *lon1*, *azi1*, *s12*, and *a12* are
     * always included in the returned result.  The value of *lon2* returned
     * is in the range [180, 180], unless the *outmask*
     * includes the [GeodesicMask.LONG_UNROLL] flag.
     */
    fun Direct(
        lat1: Double, lon1: Double,
        azi1: Double, s12: Double, outmask: Int
    ): GeodesicData {
        return Direct(lat1, lon1, azi1, false, s12, outmask)
    }

    /**
     * Solve the direct geodesic problem where the length of the geodesic
     * is specified in terms of arc length.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param a12 arc length between point 1 and point 2 (degrees); it can
     * be negative.
     * @return a [GeodesicData] object with the following fields:
     * *lat1*, *lon1*, *azi1*, *lat2*, *lon2*,
     * *azi2*, *s12*, *a12*.
     *
     *
     * *lat1* should be in the range [90, 90].  The values
     * of *lon2* and *azi2* returned are in the range [180,
     * 180].
     *
     *
     * If either point is at a pole, the azimuth is defined by keeping the
     * longitude fixed, writing *lat* = (90  ),
     * and taking the limit   0+.  An arc length greater that
     * 180 signifies a geodesic which is not a shortest path.  (For a
     * prolate ellipsoid, an additional condition is necessary for a shortest
     * path: the longitudinal extent must not exceed of 180.)
     */
    fun ArcDirect(
        lat1: Double, lon1: Double,
        azi1: Double, a12: Double
    ): GeodesicData {
        return Direct(lat1, lon1, azi1, true, a12, GeodesicMask.STANDARD)
    }

    /**
     * Solve the direct geodesic problem where the length of the geodesic is
     * specified in terms of arc length and with a subset of the geodesic results
     * returned.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param a12 arc length between point 1 and point 2 (degrees); it can
     * be negative.
     * @param outmask a bitor'ed combination of [GeodesicMask] values
     * specifying which results should be returned.
     * @return a [GeodesicData] object with the fields specified by
     * *outmask* computed.
     *
     *
     * *lat1*, *lon1*, *azi1*, and *a12* are always included
     * in the returned result.  The value of *lon2* returned is in the range
     * [180, 180], unless the *outmask* includes the [ ][GeodesicMask.LONG_UNROLL] flag.
     */
    fun ArcDirect(
        lat1: Double, lon1: Double,
        azi1: Double, a12: Double, outmask: Int
    ): GeodesicData {
        return Direct(lat1, lon1, azi1, true, a12, outmask)
    }

    /**
     * The general direct geodesic problem.  [Direct][.Direct] and
     * [ArcDirect][.ArcDirect] are defined in terms of this function.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param arcmode boolean flag determining the meaning of the
     * *s12_a12*.
     * @param s12_a12 if *arcmode* is false, this is the distance between
     * point 1 and point 2 (meters); otherwise it is the arc length between
     * point 1 and point 2 (degrees); it can be negative.
     * @param outmask a bitor'ed combination of [GeodesicMask] values
     * specifying which results should be returned.
     * @return a [GeodesicData] object with the fields specified by
     * *outmask* computed.
     *
     *
     * The [GeodesicMask] values possible for *outmask* are
     *
     *  *
     * *outmask* |= [GeodesicMask.LATITUDE] for the latitude
     * *lat2*;
     *  *
     * *outmask* |= [GeodesicMask.LONGITUDE] for the latitude
     * *lon2*;
     *  *
     * *outmask* |= [GeodesicMask.AZIMUTH] for the latitude
     * *azi2*;
     *  *
     * *outmask* |= [GeodesicMask.DISTANCE] for the distance
     * *s12*;
     *  *
     * *outmask* |= [GeodesicMask.REDUCEDLENGTH] for the reduced
     * length *m12*;
     *  *
     * *outmask* |= [GeodesicMask.GEODESICSCALE] for the geodesic
     * scales *M12* and *M21*;
     *  *
     * *outmask* |= [GeodesicMask.AREA] for the area *S12*;
     *  *
     * *outmask* |= [GeodesicMask.ALL] for all of the above;
     *  *
     * *outmask* |= [GeodesicMask.LONG_UNROLL], if set then
     * *lon1* is unchanged and *lon2*  *lon1* indicates
     * how many times and in what sense the geodesic encircles the ellipsoid.
     * Otherwise *lon1* and *lon2* are both reduced to the range
     * [180, 180].
     *
     *
     *
     * The function value *a12* is always computed and returned and this
     * equals *s12_a12* is *arcmode* is true.  If *outmask*
     * includes [GeodesicMask.DISTANCE] and *arcmode* is false, then
     * *s12* = *s12_a12*.  It is not necessary to include [ ][GeodesicMask.DISTANCE_IN] in *outmask*; this is automatically
     * included is *arcmode* is false.
     */
    fun Direct(
        lat1: Double, lon1: Double, azi1: Double,
        arcmode: Boolean, s12_a12: Double, outmask: Int
    ): GeodesicData {
        // Automatically supply DISTANCE_IN if necessary
        var _outmask = outmask
        if (!arcmode) _outmask = _outmask or GeodesicMask.DISTANCE_IN
        return GeodesicLine(this, lat1, lon1, azi1, _outmask)
            .Position(arcmode, s12_a12, _outmask) // Note the dot!
    }
    /**
     * Define a [GeodesicLine] in terms of the direct geodesic problem
     * specified in terms of distance with a subset of the capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param s12 distance between point 1 and point 2 (meters); it can be
     * negative.
     * @param caps bitor'ed combination of [GeodesicMask] values
     * specifying the capabilities the GeodesicLine object should possess,
     * i.e., which quantities can be returned in calls to
     * [GeodesicLine.Position].
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the direct geodesic problem.
     *
     *
     * *lat1* should be in the range [90, 90].
     */
    /**
     * Define a [GeodesicLine] in terms of the direct geodesic problem
     * specified in terms of distance with all capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param s12 distance between point 1 and point 2 (meters); it can be
     * negative.
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the direct geodesic problem.
     *
     *
     * *lat1* should be in the range [90, 90].
     */
    @JvmOverloads
    fun DirectLine(
        lat1: Double, lon1: Double, azi1: Double,
        s12: Double, caps: Int = GeodesicMask.ALL
    ): GeodesicLine {
        return GenDirectLine(lat1, lon1, azi1, false, s12, caps)
    }
    /**
     * Define a [GeodesicLine] in terms of the direct geodesic problem
     * specified in terms of arc length with a subset of the capabilities
     * included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param a12 arc length between point 1 and point 2 (degrees); it can
     * be negative.
     * @param caps bitor'ed combination of [GeodesicMask] values
     * specifying the capabilities the GeodesicLine object should possess,
     * i.e., which quantities can be returned in calls to
     * [GeodesicLine.Position].
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the direct geodesic problem.
     *
     *
     * *lat1* should be in the range [90, 90].
     */
    /**
     * Define a [GeodesicLine] in terms of the direct geodesic problem
     * specified in terms of arc length with all capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param a12 arc length between point 1 and point 2 (degrees); it can
     * be negative.
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the direct geodesic problem.
     *
     *
     * *lat1* should be in the range [90, 90].
     */
    @JvmOverloads
    fun ArcDirectLine(
        lat1: Double, lon1: Double, azi1: Double,
        a12: Double, caps: Int = GeodesicMask.ALL
    ): GeodesicLine {
        return GenDirectLine(lat1, lon1, azi1, true, a12, caps)
    }

    /**
     * Define a [GeodesicLine] in terms of the direct geodesic problem
     * specified in terms of either distance or arc length with a subset of the
     * capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param arcmode boolean flag determining the meaning of the *s12_a12*.
     * @param s12_a12 if *arcmode* is false, this is the distance between
     * point 1 and point 2 (meters); otherwise it is the arc length between
     * point 1 and point 2 (degrees); it can be negative.
     * @param caps bitor'ed combination of [GeodesicMask] values
     * specifying the capabilities the GeodesicLine object should possess,
     * i.e., which quantities can be returned in calls to
     * [GeodesicLine.Position].
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the direct geodesic problem.
     *
     *
     * *lat1* should be in the range [90, 90].
     */
    fun GenDirectLine(
        lat1: Double, lon1: Double, azi1: Double,
        arcmode: Boolean, s12_a12: Double, caps: Int
    ): GeodesicLine {
        var _azi1 = azi1
        var _caps = caps
        _azi1 = GeoMath.AngNormalize(_azi1)
        val salp1: Double
        val calp1: Double
        // Guard against underflow in salp0.  Also -0 is converted to +0.
        val p = Pair()
        GeoMath.sincosd(p, GeoMath.AngRound(_azi1))
        salp1 = p.first
        calp1 = p.second
        // Automatically supply DISTANCE_IN if necessary
        if (!arcmode) _caps = _caps or GeodesicMask.DISTANCE_IN
        return GeodesicLine(
            this, lat1, lon1, _azi1, salp1, calp1,
            _caps, arcmode, s12_a12
        )
    }

    private inner class InverseData {
        var g: GeodesicData
        var salp1: Double
        var calp1: Double
        var salp2: Double
        var calp2: Double

        init {
            g = GeodesicData()
            calp2 = Double.NaN
            salp2 = calp2
            calp1 = salp2
            salp1 = calp1
        }
    }

    private fun InverseInt(
        lat1: Double, lon1: Double,
        lat2: Double, lon2: Double, outmask: Int
    ): InverseData {
        var _lat1 = lat1
        var _lat2 = lat2
        val result = InverseData()
        val p = Pair()
        val r = result.g
        // Compute longitude difference (AngDiff does this carefully).  Result is
        // in [-180, 180] but -180 is only for west-going geodesics.  180 is for
        // east-going and meridional geodesics.
        _lat1 = GeoMath.LatFix(_lat1)
        r.lat1 = _lat1
        _lat2 = GeoMath.LatFix(_lat2)
        r.lat2 = _lat2
        // If really close to the equator, treat as on equator.
        _lat1 = GeoMath.AngRound(_lat1)
        _lat2 = GeoMath.AngRound(_lat2)
        var lon12: Double
        var lon12s: Double
        GeoMath.AngDiff(p, lon1, lon2)
        lon12 = p.first
        lon12s = p.second
        if (outmask and GeodesicMask.LONG_UNROLL != 0) {
            r.lon1 = lon1
            r.lon2 = lon1 + lon12 + lon12s
        } else {
            r.lon1 = GeoMath.AngNormalize(lon1)
            r.lon2 = GeoMath.AngNormalize(lon2)
        }
        // Make longitude difference positive.
        var lonsign = Math.copySign(1.0, lon12).toInt()
        lon12 *= lonsign.toDouble()
        lon12s *= lonsign.toDouble()
        val lam12 = Math.toRadians(lon12)
        val slam12: Double
        val clam12: Double
        // Calculate sincos of lon12 + error (this applies AngRound internally).
        GeoMath.sincosde(p, lon12, lon12s)
        slam12 = p.first
        clam12 = p.second
        lon12s = 180 - lon12 - lon12s // the supplementary longitude difference

        // Swap points so that point with higher (abs) latitude is point 1
        // If one latitude is a nan, then it becomes lat1.
        val swapp = if (Math.abs(_lat1) < Math.abs(_lat2) || _lat2 != _lat2) -1 else 1
        if (swapp < 0) {
            lonsign *= -1
            run {
                val t = _lat1
                _lat1 = _lat2
                _lat2 = t
            }
        }
        // Make lat1 <= 0
        val latsign = Math.copySign(1.0, -_lat1).toInt()
        _lat1 *= latsign.toDouble()
        _lat2 *= latsign.toDouble()
        // Now we have
        //
        //     0 <= lon12 <= 180
        //     -90 <= lat1 <= 0
        //     lat1 <= lat2 <= -lat1
        //
        // longsign, swapp, latsign register the transformation to bring the
        // coordinates to this canonical form.  In all cases, 1 means no change was
        // made.  We make these transformations so that there are few cases to
        // check, e.g., on verifying quadrants in atan2.  In addition, this
        // enforces some symmetries in the results returned.
        var sbet1: Double
        var cbet1: Double
        var sbet2: Double
        var cbet2: Double
        var s12x: Double
        var m12x: Double
        m12x = Double.NaN
        s12x = m12x
        GeoMath.sincosd(p, _lat1)
        sbet1 = _f1 * p.first
        cbet1 = p.second
        // Ensure cbet1 = +epsilon at poles; doing the fix on beta means that sig12
        // will be <= 2*tiny for two points at the same pole.
        GeoMath.norm(p, sbet1, cbet1)
        sbet1 = p.first
        cbet1 = p.second
        cbet1 = Math.max(tiny_, cbet1)
        GeoMath.sincosd(p, _lat2)
        sbet2 = _f1 * p.first
        cbet2 = p.second
        // Ensure cbet2 = +epsilon at poles
        GeoMath.norm(p, sbet2, cbet2)
        sbet2 = p.first
        cbet2 = p.second
        cbet2 = Math.max(tiny_, cbet2)

        // If cbet1 < -sbet1, then cbet2 - cbet1 is a sensitive measure of the
        // |bet1| - |bet2|.  Alternatively (cbet1 >= -sbet1), abs(sbet2) + sbet1 is
        // a better measure.  This logic is used in assigning calp2 in Lambda12.
        // Sometimes these quantities vanish and in that case we force bet2 = +/-
        // bet1 exactly.  An example where is is necessary is the inverse problem
        // 48.522876735459 0 -48.52287673545898293 179.599720456223079643
        // which failed with Visual Studio 10 (Release and Debug)
        if (cbet1 < -sbet1) {
            if (cbet2 == cbet1) sbet2 = Math.copySign(sbet1, sbet2)
        } else {
            if (Math.abs(sbet2) == -sbet1) cbet2 = cbet1
        }
        val dn1 = Math.sqrt(1 + _ep2 * GeoMath.sq(sbet1))
        val dn2 = Math.sqrt(1 + _ep2 * GeoMath.sq(sbet2))
        var a12: Double
        var sig12: Double
        var calp1: Double
        var salp1: Double
        var calp2: Double
        var salp2: Double
        salp2 = Double.NaN
        calp2 = salp2
        salp1 = calp2
        calp1 = salp1
        sig12 = calp1
        a12 = sig12
        // index zero elements of these arrays are unused
        val C1a = DoubleArray(nC1_ + 1)
        val C2a = DoubleArray(nC2_ + 1)
        val C3a = DoubleArray(nC3_)
        var meridian = _lat1 == -90.0 || slam12 == 0.0
        val v = LengthsV()
        if (meridian) {

            // Endpoints are on a single full meridian, so the geodesic might lie on
            // a meridian.
            calp1 = clam12
            salp1 = slam12 // Head to the target longitude
            calp2 = 1.0
            salp2 = 0.0 // At the target we're heading north
            val   // tan(bet) = tan(sig) * cos(alp)
                    ssig1 = sbet1
            val csig1 = calp1 * cbet1
            val ssig2 = sbet2
            val csig2 = calp2 * cbet2

            // sig12 = sig2 - sig1 (N.B. with Java max(+0.0, -0.0) -> +0.0)
            sig12 = Math.atan2(
                Math.max(0.0, csig1 * ssig2 - ssig1 * csig2),
                csig1 * csig2 + ssig1 * ssig2
            )
            Lengths(
                v, _n, sig12, ssig1, csig1, dn1,
                ssig2, csig2, dn2, cbet1, cbet2,
                outmask or GeodesicMask.DISTANCE or GeodesicMask.REDUCEDLENGTH,
                C1a, C2a
            )
            s12x = v.s12b
            m12x = v.m12b
            if (outmask and GeodesicMask.GEODESICSCALE != 0) {
                r.M12 = v.M12
                r.M21 = v.M21
            }
            // Add the check for sig12 since zero length geodesics might yield m12 <
            // 0.  Test case was
            //
            //    echo 20.001 0 20.001 0 | GeodSolve -i
            //
            // In fact, we will have sig12 > pi/2 for meridional geodesic which is
            // not a shortest path.
            if (sig12 < 1 || m12x >= 0) {
                // Need at least 2, to handle 90 0 90 180
                if (sig12 < 3 * tiny_ || sig12 < tol0_ && (s12x < 0 || m12x < 0)) {
                    s12x = 0.0
                    m12x = s12x
                    sig12 = m12x
                }
                m12x *= _b
                s12x *= _b
                a12 = Math.toDegrees(sig12)
            } else  // m12 < 0, i.e., prolate and too close to anti-podal
                meridian = false
        }
        var omg12 = Double.NaN
        var somg12 = 2.0
        var comg12 = Double.NaN
        if (!meridian && sbet1 == 0.0 &&  // and sbet2 == 0
            // Mimic the way Lambda12 works with calp1 = 0
            (_f <= 0 || lon12s >= _f * 180)
        ) {

            // Geodesic runs along equator
            calp2 = 0.0
            calp1 = calp2
            salp2 = 1.0
            salp1 = salp2
            s12x = _a * lam12
            omg12 = lam12 / _f1
            sig12 = omg12
            m12x = _b * Math.sin(sig12)
            if (outmask and GeodesicMask.GEODESICSCALE != 0) {
                r.M21 = Math.cos(sig12)
                r.M12 = r.M21
            }
            a12 = lon12 / _f1
        } else if (!meridian) {

            // Now point1 and point2 belong within a hemisphere bounded by a
            // meridian and geodesic is neither meridional or equatorial.

            // Figure a starting point for Newton's method
            var dnm: Double
            run {
                val s = InverseStart(
                    sbet1, cbet1, dn1, sbet2, cbet2, dn2,
                    lam12, slam12, clam12,
                    C1a, C2a, p, v
                )
                sig12 = s.sig12
                salp1 = s.salp1
                calp1 = s.calp1
                salp2 = s.salp2
                calp2 = s.calp2
                dnm = s.dnm
            }
            if (sig12 >= 0) {
                // Short lines (InverseStart sets salp2, calp2, dnm)
                s12x = sig12 * _b * dnm
                m12x = GeoMath.sq(dnm) * _b * Math.sin(sig12 / dnm)
                if (outmask and GeodesicMask.GEODESICSCALE != 0) {
                    r.M21 = Math.cos(sig12 / dnm)
                    r.M12 = r.M21
                }
                a12 = Math.toDegrees(sig12)
                omg12 = lam12 / (_f1 * dnm)
            } else {

                // Newton's method.  This is a straightforward solution of f(alp1) =
                // lambda12(alp1) - lam12 = 0 with one wrinkle.  f(alp) has exactly one
                // root in the interval (0, pi) and its derivative is positive at the
                // root.  Thus f(alp) is positive for alp > alp1 and negative for alp <
                // alp1.  During the course of the iteration, a range (alp1a, alp1b) is
                // maintained which brackets the root and with each evaluation of
                // f(alp) the range is shrunk, if possible.  Newton's method is
                // restarted whenever the derivative of f is negative (because the new
                // value of alp1 is then further from the solution) or if the new
                // estimate of alp1 lies outside (0,pi); in this case, the new starting
                // guess is taken to be (alp1a + alp1b) / 2.
                var ssig1: Double
                var csig1: Double
                var ssig2: Double
                var csig2: Double
                var eps: Double
                var domg12: Double
                domg12 = Double.NaN
                eps = domg12
                csig2 = eps
                ssig2 = csig2
                csig1 = ssig2
                ssig1 = csig1
                var numit = 0
                // Bracketing range
                var salp1a = tiny_
                var calp1a = 1.0
                var salp1b = tiny_
                var calp1b = -1.0
                val w = Lambda12V()
                var tripn = false
                var tripb = false
                while (true) {

                    // the WGS84 test set: mean = 1.47, sd = 1.25, max = 16
                    // WGS84 and random input: mean = 2.85, sd = 0.60
                    var V: Double
                    var dV: Double
                    Lambda12(
                        w, sbet1, cbet1, dn1, sbet2, cbet2, dn2, salp1, calp1,
                        slam12, clam12, numit < maxit1_, C1a, C2a, C3a, p, v
                    )
                    V = w.lam12
                    salp2 = w.salp2
                    calp2 = w.calp2
                    sig12 = w.sig12
                    ssig1 = w.ssig1
                    csig1 = w.csig1
                    ssig2 = w.ssig2
                    csig2 = w.csig2
                    eps = w.eps
                    domg12 = w.domg12
                    dV = w.dlam12
                    if (tripb ||
                        // Reversed test to allow escape with NaNs
                        // TODO: Automatic conversion was wrong !!!
                        !(Math.abs(V) >= (if (tripn) 8 else 1) * tol0_) ||
                        // Enough bisections to get accurate result
                        numit == maxit2_
                    ) break
                    // Update bracketing values
                    if (V > 0 && (numit > maxit1_ || calp1 / salp1 > calp1b / salp1b)) {
                        salp1b = salp1
                        calp1b = calp1
                    } else if (V < 0 && (numit > maxit1_ || calp1 / salp1 < calp1a / salp1a)) {
                        salp1a = salp1
                        calp1a = calp1
                    }
                    if (numit < maxit1_ && dV > 0) {
                        val dalp1 = -V / dV
                        if (Math.abs(dalp1) < Math.PI) {
                            val sdalp1 = Math.sin(dalp1)
                            val cdalp1 = Math.cos(dalp1)
                            val nsalp1 = salp1 * cdalp1 + calp1 * sdalp1
                            if (nsalp1 > 0) {
                                calp1 = calp1 * cdalp1 - salp1 * sdalp1
                                salp1 = nsalp1
                                GeoMath.norm(p, salp1, calp1)
                                salp1 = p.first
                                calp1 = p.second
                                // In some regimes we don't get quadratic convergence because
                                // slope -> 0.  So use convergence conditions based on epsilon
                                // instead of sqrt(epsilon).
                                tripn = Math.abs(V) <= 16 * tol0_
                                ++numit
                                continue
                            }
                        }
                    }
                    // Either dV was not positive or updated value was outside legal
                    // range.  Use the midpoint of the bracket as the next estimate.
                    // This mechanism is not needed for the WGS84 ellipsoid, but it does
                    // catch problems with more eccentric ellipsoids.  Its efficacy is
                    // such for the WGS84 test set with the starting guess set to alp1 =
                    // 90deg:
                    // the WGS84 test set: mean = 5.21, sd = 3.93, max = 24
                    // WGS84 and random input: mean = 4.74, sd = 0.99
                    salp1 = (salp1a + salp1b) / 2
                    calp1 = (calp1a + calp1b) / 2
                    GeoMath.norm(p, salp1, calp1)
                    salp1 = p.first
                    calp1 = p.second
                    tripn = false
                    tripb = Math.abs(salp1a - salp1) + (calp1a - calp1) < tolb_ || Math.abs(salp1 - salp1b) + (calp1 - calp1b) < tolb_
                    ++numit
                }
                run {

                    // Ensure that the reduced length and geodesic scale are computed in
                    // a "canonical" way, with the I2 integral.
                    val lengthmask = outmask or
                            if (outmask and
                                (GeodesicMask.REDUCEDLENGTH or GeodesicMask.GEODESICSCALE) != 0
                            ) GeodesicMask.DISTANCE else GeodesicMask.NONE
                    Lengths(
                        v, eps, sig12,
                        ssig1, csig1, dn1, ssig2, csig2, dn2, cbet1, cbet2,
                        lengthmask, C1a, C2a
                    )
                    s12x = v.s12b
                    m12x = v.m12b
                    if (outmask and GeodesicMask.GEODESICSCALE != 0) {
                        r.M12 = v.M12
                        r.M21 = v.M21
                    }
                }
                m12x *= _b
                s12x *= _b
                a12 = Math.toDegrees(sig12)
                if (outmask and GeodesicMask.AREA != 0) {
                    // omg12 = lam12 - domg12
                    val sdomg12 = Math.sin(domg12)
                    val cdomg12 = Math.cos(domg12)
                    somg12 = slam12 * cdomg12 - clam12 * sdomg12
                    comg12 = clam12 * cdomg12 + slam12 * sdomg12
                }
            }
        }
        if (outmask and GeodesicMask.DISTANCE != 0) r.s12 = 0 + s12x // Convert -0 to 0
        if (outmask and GeodesicMask.REDUCEDLENGTH != 0) r.m12 = 0 + m12x // Convert -0 to 0
        if (outmask and GeodesicMask.AREA != 0) {
            val   // From Lambda12: sin(alp1) * cos(bet1) = sin(alp0)
                    salp0 = salp1 * cbet1
            val calp0 = Math.hypot(calp1, salp1 * sbet1) // calp0 > 0
            val alp12: Double
            if (calp0 != 0.0 && salp0 != 0.0) {
                var   // From Lambda12: tan(bet) = tan(sig) * cos(alp)
                        ssig1 = sbet1
                var csig1 = calp1 * cbet1
                var ssig2 = sbet2
                var csig2 = calp2 * cbet2
                val k2 = GeoMath.sq(calp0) * _ep2
                val eps = k2 / (2 * (1 + Math.sqrt(1 + k2)) + k2)
                // Multiplier = a^2 * e^2 * cos(alpha0) * sin(alpha0).
                val A4 = GeoMath.sq(_a) * calp0 * salp0 * _e2
                GeoMath.norm(p, ssig1, csig1)
                ssig1 = p.first
                csig1 = p.second
                GeoMath.norm(p, ssig2, csig2)
                ssig2 = p.first
                csig2 = p.second
                val C4a = DoubleArray(nC4_)
                C4f(eps, C4a)
                val B41 = SinCosSeries(false, ssig1, csig1, C4a)
                val B42 = SinCosSeries(false, ssig2, csig2, C4a)
                r.S12 = A4 * (B42 - B41)
            } else  // Avoid problems with indeterminate sig1, sig2 on equator
                r.S12 = 0.0
            if (!meridian && somg12 == 2.0) {
                somg12 = Math.sin(omg12)
                comg12 = Math.cos(omg12)
            }
            if (!meridian && comg12 > -0.7071 && // Long difference not too big
                sbet2 - sbet1 < 1.75
            ) { // Lat difference not too big
                // Use tan(Gamma/2) = tan(omg12/2)
                // * (tan(bet1/2)+tan(bet2/2))/(1+tan(bet1/2)*tan(bet2/2))
                // with tan(x/2) = sin(x)/(1+cos(x))
                val domg12 = 1 + comg12
                val dbet1 = 1 + cbet1
                val dbet2 = 1 + cbet2
                alp12 = 2 * Math.atan2(
                    somg12 * (sbet1 * dbet2 + sbet2 * dbet1),
                    domg12 * (sbet1 * sbet2 + dbet1 * dbet2)
                )
            } else {
                // alp12 = alp2 - alp1, used in atan2 so no need to normalize
                var salp12 = salp2 * calp1 - calp2 * salp1
                var calp12 = calp2 * calp1 + salp2 * salp1
                // The right thing appears to happen if alp1 = +/-180 and alp2 = 0, viz
                // salp12 = -0 and alp12 = -180.  However this depends on the sign
                // being attached to 0 correctly.  The following ensures the correct
                // behavior.
                if (salp12 == 0.0 && calp12 < 0) {
                    salp12 = tiny_ * calp1
                    calp12 = -1.0
                }
                alp12 = Math.atan2(salp12, calp12)
            }
            r.S12 += _c2 * alp12
            r.S12 *= (swapp * lonsign * latsign).toDouble()
            // Convert -0 to 0
            r.S12 += 0.0
        }

        // Convert calp, salp to azimuth accounting for lonsign, swapp, latsign.
        if (swapp < 0) {
            run {
                val t = salp1
                salp1 = salp2
                salp2 = t
            }
            run {
                val t = calp1
                calp1 = calp2
                calp2 = t
            }
            if (outmask and GeodesicMask.GEODESICSCALE != 0) {
                val t = r.M12
                r.M12 = r.M21
                r.M21 = t
            }
        }
        salp1 *= (swapp * lonsign).toDouble()
        calp1 *= (swapp * latsign).toDouble()
        salp2 *= (swapp * lonsign).toDouble()
        calp2 *= (swapp * latsign).toDouble()

        // Returned value in [0, 180]
        r.a12 = a12
        result.salp1 = salp1
        result.calp1 = calp1
        result.salp2 = salp2
        result.calp2 = calp2
        return result
    }
    /**
     * Solve the inverse geodesic problem with a subset of the geodesic results
     * returned.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param lat2 latitude of point 2 (degrees).
     * @param lon2 longitude of point 2 (degrees).
     * @param outmask a bitor'ed combination of [GeodesicMask] values
     * specifying which results should be returned.
     * @return a [GeodesicData] object with the fields specified by
     * *outmask* computed.
     *
     *
     * The [GeodesicMask] values possible for *outmask* are
     *
     *  *
     * *outmask* |= [GeodesicMask.DISTANCE] for the distance
     * *s12*;
     *  *
     * *outmask* |= [GeodesicMask.AZIMUTH] for the latitude
     * *azi2*;
     *  *
     * *outmask* |= [GeodesicMask.REDUCEDLENGTH] for the reduced
     * length *m12*;
     *  *
     * *outmask* |= [GeodesicMask.GEODESICSCALE] for the geodesic
     * scales *M12* and *M21*;
     *  *
     * *outmask* |= [GeodesicMask.AREA] for the area *S12*;
     *  *
     * *outmask* |= [GeodesicMask.ALL] for all of the above.
     *  *
     * *outmask* |= [GeodesicMask.LONG_UNROLL], if set then
     * *lon1* is unchanged and *lon2*  *lon1* indicates
     * whether the geodesic is east going or west going.  Otherwise *lon1*
     * and *lon2* are both reduced to the range [180,
     * 180].
     *
     *
     *
     * *lat1*, *lon1*, *lat2*, *lon2*, and *a12* are
     * always included in the returned result.
     */
    /**
     * Solve the inverse geodesic problem.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param lat2 latitude of point 2 (degrees).
     * @param lon2 longitude of point 2 (degrees).
     * @return a [GeodesicData] object with the following fields:
     * *lat1*, *lon1*, *azi1*, *lat2*, *lon2*,
     * *azi2*, *s12*, *a12*.
     *
     *
     * *lat1* and *lat2* should be in the range [90,
     * 90].  The values of *azi1* and *azi2* returned are in the
     * range [180, 180].
     *
     *
     * If either point is at a pole, the azimuth is defined by keeping the
     * longitude fixed, writing *lat* = (90  ),
     * taking the limit   0+.
     *
     *
     * The solution to the inverse problem is found using Newton's method.  If
     * this fails to converge (this is very unlikely in geodetic applications
     * but does occur for very eccentric ellipsoids), then the bisection method
     * is used to refine the solution.
     */
    @JvmOverloads
    fun Inverse(
        lat1: Double, lon1: Double,
        lat2: Double, lon2: Double, outmask: Int = GeodesicMask.STANDARD
    ): GeodesicData {
        var _outmask = outmask
        _outmask = _outmask and GeodesicMask.OUT_MASK
        val result = InverseInt(lat1, lon1, lat2, lon2, _outmask)
        val r = result.g
        if (_outmask and GeodesicMask.AZIMUTH != 0) {
            r.azi1 = GeoMath.atan2d(result.salp1, result.calp1)
            r.azi2 = GeoMath.atan2d(result.salp2, result.calp2)
        }
        return r
    }
    /**
     * Define a [GeodesicLine] in terms of the inverse geodesic problem
     * with a subset of the capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param lat2 latitude of point 2 (degrees).
     * @param lon2 longitude of point 2 (degrees).
     * @param caps bitor'ed combination of [GeodesicMask] values specifying
     * the capabilities the GeodesicLine object should possess, i.e., which
     * quantities can be returned in calls to
     * [GeodesicLine.Position].
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the inverse geodesic problem.
     *
     *
     * *lat1* and *lat2* should be in the range [90,
     * 90].
     */
    /**
     * Define a [GeodesicLine] in terms of the inverse geodesic problem
     * with all capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param lat2 latitude of point 2 (degrees).
     * @param lon2 longitude of point 2 (degrees).
     * @return a [GeodesicLine] object.
     *
     *
     * This function sets point 3 of the GeodesicLine to correspond to point 2
     * of the inverse geodesic problem.
     *
     *
     * *lat1* and *lat2* should be in the range [90,
     * 90].
     */
    @JvmOverloads
    fun InverseLine(
        lat1: Double, lon1: Double,
        lat2: Double, lon2: Double, caps: Int = GeodesicMask.ALL
    ): GeodesicLine {
        var _caps = caps
        val result = InverseInt(lat1, lon1, lat2, lon2, 0)
        val salp1 = result.salp1
        val calp1 = result.calp1
        val azi1 = GeoMath.atan2d(salp1, calp1)
        val a12 = result.g.a12
        // Ensure that a12 can be converted to a distance
        if (_caps and (GeodesicMask.OUT_MASK and GeodesicMask.DISTANCE_IN) != 0) _caps = _caps or GeodesicMask.DISTANCE
        return GeodesicLine(
            this, lat1, lon1, azi1, salp1, calp1, _caps,
            true, a12
        )
    }
    /**
     * Set up to compute several points on a single geodesic with a subset of the
     * capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @param caps bitor'ed combination of [GeodesicMask] values specifying
     * the capabilities the [GeodesicLine] object should possess, i.e.,
     * which quantities can be returned in calls to [   ][GeodesicLine.Position].
     * @return a [GeodesicLine] object.
     *
     *
     * The [GeodesicMask] values are
     *
     *  *
     * *caps* |= [GeodesicMask.LATITUDE] for the latitude
     * *lat2*; this is added automatically;
     *  *
     * *caps* |= [GeodesicMask.LONGITUDE] for the latitude
     * *lon2*;
     *  *
     * *caps* |= [GeodesicMask.AZIMUTH] for the azimuth *azi2*;
     * this is added automatically;
     *  *
     * *caps* |= [GeodesicMask.DISTANCE] for the distance
     * *s12*;
     *  *
     * *caps* |= [GeodesicMask.REDUCEDLENGTH] for the reduced length
     * *m12*;
     *  *
     * *caps* |= [GeodesicMask.GEODESICSCALE] for the geodesic
     * scales *M12* and *M21*;
     *  *
     * *caps* |= [GeodesicMask.AREA] for the area *S12*;
     *  *
     * *caps* |= [GeodesicMask.DISTANCE_IN] permits the length of
     * the geodesic to be given in terms of *s12*; without this capability
     * the length can only be specified in terms of arc length;
     *  *
     * *caps* |= [GeodesicMask.ALL] for all of the above.
     *
     *
     *
     * If the point is at a pole, the azimuth is defined by keeping *lon1*
     * fixed, writing *lat1* = (90  ), and taking
     * the limit   0+.
     */
    /**
     * Set up to compute several points on a single geodesic with all
     * capabilities included.
     *
     *
     * @param lat1 latitude of point 1 (degrees).
     * @param lon1 longitude of point 1 (degrees).
     * @param azi1 azimuth at point 1 (degrees).
     * @return a [GeodesicLine] object.
     *
     *
     * *lat1* should be in the range [90, 90].  The full
     * set of capabilities is included.
     *
     *
     * If the point is at a pole, the azimuth is defined by keeping the
     * *lon1* fixed, writing *lat1* = (90  ),
     * taking the limit   0+.
     */
    @JvmOverloads
    fun Line(lat1: Double, lon1: Double, azi1: Double, caps: Int = GeodesicMask.ALL): GeodesicLine {
        return GeodesicLine(this, lat1, lon1, azi1, caps)
    }

    /**
     * @return *a* the equatorial radius of the ellipsoid (meters).  This is
     * the value used in the constructor.
     */
    fun EquatorialRadius(): Double {
        return _a
    }

    /**
     * @return *f* the  flattening of the ellipsoid.  This is the
     * value used in the constructor.
     */
    fun Flattening(): Double {
        return _f
    }

    /**
     * @return total area of ellipsoid in meters<sup>2</sup>.  The area of a
     * polygon encircling a pole can be found by adding EllipsoidArea()/2 to
     * the sum of *S12* for each side of the polygon.
     */
    fun EllipsoidArea(): Double {
        return 4 * Math.PI * _c2
    }

    /**
     * Constructor for an ellipsoid with
     *
     *
     * @param a equatorial radius (meters).
     * @param f flattening of ellipsoid.  Setting *f* = 0 gives a sphere.
     * Negative *f* gives a prolate ellipsoid.
     * @exception GeographicErr if *a* or (1  *f* ) *a* is
     * not positive.
     */
    init {
        _f1 = 1 - _f
        _e2 = _f * (2 - _f)
        _ep2 = _e2 / GeoMath.sq(_f1) // e2 / (1 - e2)
        _n = _f / (2 - _f)
        _b = _a * _f1
        _c2 = (GeoMath.sq(_a) + GeoMath.sq(_b) *
                if (_e2 == 0.0) 1.0 else (if (_e2 > 0) GeoMath.atanh(Math.sqrt(_e2)) else Math.atan(Math.sqrt(-_e2))) /
                        Math.sqrt(Math.abs(_e2))) / 2 // authalic radius squared
        // The sig12 threshold for "really short".  Using the auxiliary sphere
        // solution with dnm computed at (bet1 + bet2) / 2, the relative error in
        // the azimuth consistency check is sig12^2 * abs(f) * min(1, 1-f/2) / 2.
        // (Error measured for 1/100 < b/a < 100 and abs(f) >= 1/1000.  For a
        // given f and sig12, the max error occurs for lines near the pole.  If
        // the old rule for computing dnm = (dn1 + dn2)/2 is used, then the error
        // increases by a factor of 2.)  Setting this equal to epsilon gives
        // sig12 = etol2.  Here 0.1 is a safety factor (error decreased by 100)
        // and max(0.001, abs(f)) stops etol2 getting too large in the nearly
        // spherical case.
        _etol2 = 0.1 * tol2_ /
                Math.sqrt(
                    Math.max(0.001, Math.abs(_f)) *
                            Math.min(1.0, 1 - _f / 2) / 2
                )
        if (!(Math.isFinite(_a) && _a > 0)) throw GeographicErr("Equatorial radius is not positive")
        if (!(Math.isFinite(_b) && _b > 0)) throw GeographicErr("Polar semi-axis is not positive")
        _A3x = DoubleArray(nA3x_)
        _C3x = DoubleArray(nC3x_)
        _C4x = DoubleArray(nC4x_)
        A3coeff()
        C3coeff()
        C4coeff()
    }

    private inner class LengthsV {
        var s12b: Double
        var m12b: Double
        var m0: Double
        var M12: Double
        var M21: Double

        init {
            M21 = Double.NaN
            M12 = M21
            m0 = M12
            m12b = m0
            s12b = m12b
        }
    }

    private fun Lengths(
        v: LengthsV, eps: Double, sig12: Double,
        ssig1: Double, csig1: Double, dn1: Double,
        ssig2: Double, csig2: Double, dn2: Double,
        cbet1: Double, cbet2: Double,
        outmask: Int,  // Scratch areas of the right size
        C1a: DoubleArray, C2a: DoubleArray
    ) {
        // Return m12b = (reduced length)/_b; also calculate s12b = distance/_b,
        // and m0 = coefficient of secular term in expression for reduced length.
        var _outmask = outmask
        _outmask = _outmask and GeodesicMask.OUT_MASK
        // v to hold compute s12b, m12b, m0, M12, M21;
        var m0x = 0.0
        var J12 = 0.0
        var A1 = 0.0
        var A2 = 0.0
        if (_outmask and (GeodesicMask.DISTANCE or GeodesicMask.REDUCEDLENGTH or
                    GeodesicMask.GEODESICSCALE) != 0
        ) {
            A1 = A1m1f(eps)
            C1f(eps, C1a)
            if (_outmask and (GeodesicMask.REDUCEDLENGTH or
                        GeodesicMask.GEODESICSCALE) != 0
            ) {
                A2 = A2m1f(eps)
                C2f(eps, C2a)
                m0x = A1 - A2
                A2 = 1 + A2
            }
            A1 = 1 + A1
        }
        if (_outmask and GeodesicMask.DISTANCE != 0) {
            val B1 = SinCosSeries(true, ssig2, csig2, C1a) -
                    SinCosSeries(true, ssig1, csig1, C1a)
            // Missing a factor of _b
            v.s12b = A1 * (sig12 + B1)
            if (_outmask and (GeodesicMask.REDUCEDLENGTH or
                        GeodesicMask.GEODESICSCALE) != 0
            ) {
                val B2 = SinCosSeries(true, ssig2, csig2, C2a) -
                        SinCosSeries(true, ssig1, csig1, C2a)
                J12 = m0x * sig12 + (A1 * B1 - A2 * B2)
            }
        } else if (_outmask and (GeodesicMask.REDUCEDLENGTH or
                    GeodesicMask.GEODESICSCALE) != 0
        ) {
            // Assume here that nC1_ >= nC2_
            for (l in 1..nC2_) C2a[l] = A1 * C1a[l] - A2 * C2a[l]
            J12 = m0x * sig12 + (SinCosSeries(true, ssig2, csig2, C2a) -
                    SinCosSeries(true, ssig1, csig1, C2a))
        }
        if (_outmask and GeodesicMask.REDUCEDLENGTH != 0) {
            v.m0 = m0x
            // Missing a factor of _b.
            // Add parens around (csig1 * ssig2) and (ssig1 * csig2) to ensure
            // accurate cancellation in the case of coincident points.
            v.m12b = dn2 * (csig1 * ssig2) - dn1 * (ssig1 * csig2) - csig1 * csig2 * J12
        }
        if (_outmask and GeodesicMask.GEODESICSCALE != 0) {
            val csig12 = csig1 * csig2 + ssig1 * ssig2
            val t = _ep2 * (cbet1 - cbet2) * (cbet1 + cbet2) / (dn1 + dn2)
            v.M12 = csig12 + (t * ssig2 - csig2 * J12) * ssig1 / dn1
            v.M21 = csig12 - (t * ssig1 - csig1 * J12) * ssig2 / dn2
        }
    }

    private inner class InverseStartV {
        var sig12: Double
        var salp1: Double
        var calp1: Double

        // Only updated if return val >= 0
        var salp2: Double
        var calp2: Double

        // Only updated for short lines
        var dnm: Double

        init {
            dnm = Double.NaN
            calp2 = dnm
            salp2 = calp2
            calp1 = salp2
            salp1 = calp1
            sig12 = salp1
        }
    }

    private fun InverseStart(
        sbet1: Double, cbet1: Double, dn1: Double,
        sbet2: Double, cbet2: Double, dn2: Double,
        lam12: Double,
        slam12: Double, clam12: Double,  // Scratch areas of the right size
        C1a: DoubleArray, C2a: DoubleArray,
        p: Pair, v: LengthsV
    ): InverseStartV {
        // Return a starting point for Newton's method in salp1 and calp1 (function
        // value is -1).  If Newton's method doesn't need to be used, return also
        // salp2 and calp2 and function value is sig12.

        // To hold sig12, salp1, calp1, salp2, calp2, dnm.
        val w = InverseStartV()
        w.sig12 = -1.0 // Return value
        val   // bet12 = bet2 - bet1 in [0, pi); bet12a = bet2 + bet1 in (-pi, 0]
                sbet12 = sbet2 * cbet1 - cbet2 * sbet1
        val cbet12 = cbet2 * cbet1 + sbet2 * sbet1
        val sbet12a = sbet2 * cbet1 + cbet2 * sbet1
        val shortline = cbet12 >= 0 && sbet12 < 0.5 && cbet2 * lam12 < 0.5
        var somg12: Double
        var comg12: Double
        if (shortline) {
            var sbetm2 = GeoMath.sq(sbet1 + sbet2)
            // sin((bet1+bet2)/2)^2
            // =  (sbet1 + sbet2)^2 / ((sbet1 + sbet2)^2 + (cbet1 + cbet2)^2)
            sbetm2 /= sbetm2 + GeoMath.sq(cbet1 + cbet2)
            w.dnm = Math.sqrt(1 + _ep2 * sbetm2)
            val omg12 = lam12 / (_f1 * w.dnm)
            somg12 = Math.sin(omg12)
            comg12 = Math.cos(omg12)
        } else {
            somg12 = slam12
            comg12 = clam12
        }
        w.salp1 = cbet2 * somg12
        w.calp1 = if (comg12 >= 0) sbet12 + cbet2 * sbet1 * GeoMath.sq(somg12) / (1 + comg12) else sbet12a - cbet2 * sbet1 * GeoMath.sq(somg12) / (1 - comg12)
        val ssig12 = Math.hypot(w.salp1, w.calp1)
        val csig12 = sbet1 * sbet2 + cbet1 * cbet2 * comg12
        if (shortline && ssig12 < _etol2) {
            // really short lines
            w.salp2 = cbet1 * somg12
            w.calp2 = sbet12 - cbet1 * sbet2 * if (comg12 >= 0) GeoMath.sq(somg12) / (1 + comg12) else 1 - comg12
            GeoMath.norm(p, w.salp2, w.calp2)
            w.salp2 = p.first
            w.calp2 = p.second
            // Set return value
            w.sig12 = Math.atan2(ssig12, csig12)
        } else if (Math.abs(_n) > 0.1 || // Skip astroid calc if too eccentric
            csig12 >= 0 || ssig12 >= 6 * Math.abs(_n) * Math.PI * GeoMath.sq(cbet1)
        ) {
            // Nothing to do, zeroth order spherical approximation is OK
        } else {
            // Scale lam12 and bet2 to x, y coordinate system where antipodal point
            // is at origin and singular point is at y = 0, x = -1.
            val x: Double
            val y: Double
            var lamscale: Double
            val betscale: Double
            val lam12x = Math.atan2(-slam12, -clam12) // lam12 - pi
            if (_f >= 0) {            // In fact f == 0 does not get here
                // x = dlong, y = dlat
                run {
                    val k2 = GeoMath.sq(sbet1) * _ep2
                    val eps = k2 / (2 * (1 + Math.sqrt(1 + k2)) + k2)
                    lamscale = _f * cbet1 * A3f(eps) * Math.PI
                }
                betscale = lamscale * cbet1
                x = lam12x / lamscale
                y = sbet12a / betscale
            } else {                  // _f < 0
                // x = dlat, y = dlong
                val cbet12a = cbet2 * cbet1 - sbet2 * sbet1
                val bet12a = Math.atan2(sbet12a, cbet12a)
                val m12b: Double
                val m0: Double
                // In the case of lon12 = 180, this repeats a calculation made in
                // Inverse.
                Lengths(
                    v, _n, Math.PI + bet12a,
                    sbet1, -cbet1, dn1, sbet2, cbet2, dn2,
                    cbet1, cbet2, GeodesicMask.REDUCEDLENGTH, C1a, C2a
                )
                m12b = v.m12b
                m0 = v.m0
                x = -1 + m12b / (cbet1 * cbet2 * m0 * Math.PI)
                betscale = if (x < -0.01) sbet12a / x else -_f * GeoMath.sq(cbet1) * Math.PI
                lamscale = betscale / cbet1
                y = lam12x / lamscale
            }
            if (y > -tol1_ && x > -1 - xthresh_) {
                // strip near cut
                if (_f >= 0) {
                    w.salp1 = Math.min(1.0, -x)
                    w.calp1 = -Math.sqrt(1 - GeoMath.sq(w.salp1))
                } else {
                    w.calp1 = Math.max(if (x > -tol1_) 0.0 else -1.0, x)
                    w.salp1 = Math.sqrt(1 - GeoMath.sq(w.calp1))
                }
            } else {
                // Estimate alp1, by solving the astroid problem.
                //
                // Could estimate alpha1 = theta + pi/2, directly, i.e.,
                //   calp1 = y/k; salp1 = -x/(1+k);  for _f >= 0
                //   calp1 = x/(1+k); salp1 = -y/k;  for _f < 0 (need to check)
                //
                // However, it's better to estimate omg12 from astroid and use
                // spherical formula to compute alp1.  This reduces the mean number of
                // Newton iterations for astroid cases from 2.24 (min 0, max 6) to 2.12
                // (min 0 max 5).  The changes in the number of iterations are as
                // follows:
                //
                // change percent
                //    1       5
                //    0      78
                //   -1      16
                //   -2       0.6
                //   -3       0.04
                //   -4       0.002
                //
                // The histogram of iterations is (m = number of iterations estimating
                // alp1 directly, n = number of iterations estimating via omg12, total
                // number of trials = 148605):
                //
                //  iter    m      n
                //    0   148    186
                //    1 13046  13845
                //    2 93315 102225
                //    3 36189  32341
                //    4  5396      7
                //    5   455      1
                //    6    56      0
                //
                // Because omg12 is near pi, estimate work with omg12a = pi - omg12
                val k = Astroid(x, y)
                val omg12a = lamscale * if (_f >= 0) -x * k / (1 + k) else -y * (1 + k) / k
                somg12 = Math.sin(omg12a)
                comg12 = -Math.cos(omg12a)
                // Update spherical estimate of alp1 using omg12 instead of lam12
                w.salp1 = cbet2 * somg12
                w.calp1 = sbet12a - cbet2 * sbet1 * GeoMath.sq(somg12) / (1 - comg12)
            }
        }
        // Sanity check on starting guess.  Backwards check allows NaN through.
        // TODO: Automatic conversion was wrong !!!
        if (!(w.salp1 <= 0)) {
            GeoMath.norm(p, w.salp1, w.calp1)
            w.salp1 = p.first
            w.calp1 = p.second
        } else {
            w.salp1 = 1.0
            w.calp1 = 0.0
        }
        return w
    }

    private inner class Lambda12V {
        var lam12: Double
        var salp2: Double
        var calp2: Double
        var sig12: Double
        var ssig1: Double
        var csig1: Double
        var ssig2: Double
        var csig2: Double
        var eps: Double
        var domg12: Double
        var dlam12: Double

        init {
            dlam12 = Double.NaN
            domg12 = dlam12
            eps = domg12
            csig2 = eps
            ssig2 = csig2
            csig1 = ssig2
            ssig1 = csig1
            sig12 = ssig1
            calp2 = sig12
            salp2 = calp2
            lam12 = salp2
        }
    }

    private fun Lambda12(
        w: Lambda12V,
        sbet1: Double, cbet1: Double, dn1: Double,
        sbet2: Double, cbet2: Double, dn2: Double,
        salp1: Double, calp1: Double,
        slam120: Double, clam120: Double,
        diffp: Boolean,  // Scratch areas of the right size
        C1a: DoubleArray, C2a: DoubleArray, C3a: DoubleArray,
        p: Pair, v: LengthsV
    ) {
        // Object to hold lam12, salp2, calp2, sig12, ssig1, csig1, ssig2, csig2,
        // eps, domg12, dlam12;
        var _calp1 = calp1
        if (sbet1 == 0.0 && _calp1 == 0.0) // Break degeneracy of equatorial line.  This case has already been
        // handled.
            _calp1 = -tiny_
        val   // sin(alp1) * cos(bet1) = sin(alp0)
                salp0 = salp1 * cbet1
        val calp0 = Math.hypot(_calp1, salp1 * sbet1) // calp0 > 0
        val somg1: Double
        val comg1: Double
        val somg2: Double
        val comg2: Double
        val somg12: Double
        val comg12: Double
        // tan(bet1) = tan(sig1) * cos(alp1)
        // tan(omg1) = sin(alp0) * tan(sig1) = tan(omg1)=tan(alp1)*sin(bet1)
        w.ssig1 = sbet1
        somg1 = salp0 * sbet1
        comg1 = _calp1 * cbet1
        w.csig1 = comg1
        GeoMath.norm(p, w.ssig1, w.csig1)
        w.ssig1 = p.first
        w.csig1 = p.second
        // GeoMath.norm(somg1, comg1); -- don't need to normalize!

        // Enforce symmetries in the case abs(bet2) = -bet1.  Need to be careful
        // about this case, since this can yield singularities in the Newton
        // iteration.
        // sin(alp2) * cos(bet2) = sin(alp0)
        w.salp2 = if (cbet2 != cbet1) salp0 / cbet2 else salp1
        // calp2 = sqrt(1 - sq(salp2))
        //       = sqrt(sq(calp0) - sq(sbet2)) / cbet2
        // and subst for calp0 and rearrange to give (choose positive sqrt
        // to give alp2 in [0, pi/2]).
        w.calp2 = if (cbet2 != cbet1 || Math.abs(sbet2) != -sbet1) Math.sqrt(
            GeoMath.sq(_calp1 * cbet1) +
                    if (cbet1 < -sbet1) (cbet2 - cbet1) * (cbet1 + cbet2) else (sbet1 - sbet2) * (sbet1 + sbet2)
        ) / cbet2 else Math.abs(_calp1)
        // tan(bet2) = tan(sig2) * cos(alp2)
        // tan(omg2) = sin(alp0) * tan(sig2).
        w.ssig2 = sbet2
        somg2 = salp0 * sbet2
        comg2 = w.calp2 * cbet2
        w.csig2 = comg2
        GeoMath.norm(p, w.ssig2, w.csig2)
        w.ssig2 = p.first
        w.csig2 = p.second
        // GeoMath.norm(somg2, comg2); -- don't need to normalize!

        // sig12 = sig2 - sig1, limit to [0, pi]
        w.sig12 = Math.atan2(
            Math.max(0.0, w.csig1 * w.ssig2 - w.ssig1 * w.csig2),
            w.csig1 * w.csig2 + w.ssig1 * w.ssig2
        )

        // omg12 = omg2 - omg1, limit to [0, pi]
        somg12 = Math.max(0.0, comg1 * somg2 - somg1 * comg2)
        comg12 = comg1 * comg2 + somg1 * somg2
        // eta = omg12 - lam120
        val eta = Math.atan2(
            somg12 * clam120 - comg12 * slam120,
            comg12 * clam120 + somg12 * slam120
        )
        val B312: Double
        val k2 = GeoMath.sq(calp0) * _ep2
        w.eps = k2 / (2 * (1 + Math.sqrt(1 + k2)) + k2)
        C3f(w.eps, C3a)
        B312 = SinCosSeries(true, w.ssig2, w.csig2, C3a) - SinCosSeries(true, w.ssig1, w.csig1, C3a)
        w.domg12 = -_f * A3f(w.eps) * salp0 * (w.sig12 + B312)
        w.lam12 = eta + w.domg12
        if (diffp) {
            if (w.calp2 == 0.0) w.dlam12 = -2 * _f1 * dn1 / sbet1 else {
                Lengths(
                    v, w.eps,
                    w.sig12, w.ssig1, w.csig1, dn1, w.ssig2, w.csig2, dn2,
                    cbet1, cbet2, GeodesicMask.REDUCEDLENGTH, C1a, C2a
                )
                w.dlam12 = v.m12b
                w.dlam12 *= _f1 / (w.calp2 * cbet2)
            }
        }
    }

    fun A3f(eps: Double): Double {
        // Evaluate A3
        return GeoMath.polyval(nA3_ - 1, _A3x, 0, eps)
    }

    fun C3f(eps: Double, c: DoubleArray) {
        // Evaluate C3 coeffs
        // Elements c[1] thru c[nC3_ - 1] are set
        var mult = 1.0
        var o = 0
        for (l in 1 until nC3_) { // l is index of C3[l]
            val m = nC3_ - l - 1 // order of polynomial in eps
            mult *= eps
            c[l] = mult * GeoMath.polyval(m, _C3x, o, eps)
            o += m + 1
        }
    }

    fun C4f(eps: Double, c: DoubleArray) {
        // Evaluate C4 coeffs
        // Elements c[0] thru c[nC4_ - 1] are set
        var mult = 1.0
        var o = 0
        for (l in 0 until nC4_) { // l is index of C4[l]
            val m = nC4_ - l - 1 // order of polynomial in eps
            c[l] = mult * GeoMath.polyval(m, _C4x, o, eps)
            o += m + 1
            mult *= eps
        }
    }

    // The scale factor A3 = mean value of (d/dsigma)I3
    protected fun A3coeff() {
        val coeff = doubleArrayOf( // A3, coeff of eps^5, polynomial in n of order 0
            -3.0, 128.0,  // A3, coeff of eps^4, polynomial in n of order 1
            -2.0, -3.0, 64.0,  // A3, coeff of eps^3, polynomial in n of order 2
            -1.0, -3.0, -1.0, 16.0,  // A3, coeff of eps^2, polynomial in n of order 2
            3.0, -1.0, -2.0, 8.0,  // A3, coeff of eps^1, polynomial in n of order 1
            1.0, -1.0, 2.0,  // A3, coeff of eps^0, polynomial in n of order 0
            1.0, 1.0
        )
        var o = 0
        var k = 0
        for (j in nA3_ - 1 downTo 0) { // coeff of eps^j
            val m = Math.min(nA3_ - j - 1, j) // order of polynomial in n
            _A3x[k++] = GeoMath.polyval(m, coeff, o, _n) / coeff[o + m + 1]
            o += m + 2
        }
    }

    // The coefficients C3[l] in the Fourier expansion of B3
    protected fun C3coeff() {
        val coeff = doubleArrayOf( // C3[1], coeff of eps^5, polynomial in n of order 0
            3.0, 128.0,  // C3[1], coeff of eps^4, polynomial in n of order 1
            2.0, 5.0, 128.0,  // C3[1], coeff of eps^3, polynomial in n of order 2
            -1.0, 3.0, 3.0, 64.0,  // C3[1], coeff of eps^2, polynomial in n of order 2
            -1.0, 0.0, 1.0, 8.0,  // C3[1], coeff of eps^1, polynomial in n of order 1
            -1.0, 1.0, 4.0,  // C3[2], coeff of eps^5, polynomial in n of order 0
            5.0, 256.0,  // C3[2], coeff of eps^4, polynomial in n of order 1
            1.0, 3.0, 128.0,  // C3[2], coeff of eps^3, polynomial in n of order 2
            -3.0, -2.0, 3.0, 64.0,  // C3[2], coeff of eps^2, polynomial in n of order 2
            1.0, -3.0, 2.0, 32.0,  // C3[3], coeff of eps^5, polynomial in n of order 0
            7.0, 512.0,  // C3[3], coeff of eps^4, polynomial in n of order 1
            -10.0, 9.0, 384.0,  // C3[3], coeff of eps^3, polynomial in n of order 2
            5.0, -9.0, 5.0, 192.0,  // C3[4], coeff of eps^5, polynomial in n of order 0
            7.0, 512.0,  // C3[4], coeff of eps^4, polynomial in n of order 1
            -14.0, 7.0, 512.0,  // C3[5], coeff of eps^5, polynomial in n of order 0
            21.0, 2560.0
        )
        var o = 0
        var k = 0
        for (l in 1 until nC3_) {        // l is index of C3[l]
            for (j in nC3_ - 1 downTo l) { // coeff of eps^j
                val m = Math.min(nC3_ - j - 1, j) // order of polynomial in n
                _C3x[k++] = GeoMath.polyval(m, coeff, o, _n) / coeff[o + m + 1]
                o += m + 2
            }
        }
    }

    protected fun C4coeff() {
        val coeff = doubleArrayOf( // C4[0], coeff of eps^5, polynomial in n of order 0
            97.0, 15015.0,  // C4[0], coeff of eps^4, polynomial in n of order 1
            1088.0, 156.0, 45045.0,  // C4[0], coeff of eps^3, polynomial in n of order 2
            -224.0, -4784.0, 1573.0, 45045.0,  // C4[0], coeff of eps^2, polynomial in n of order 3
            -10656.0, 14144.0, -4576.0, -858.0, 45045.0,  // C4[0], coeff of eps^1, polynomial in n of order 4
            64.0, 624.0, -4576.0, 6864.0, -3003.0, 15015.0,  // C4[0], coeff of eps^0, polynomial in n of order 5
            100.0, 208.0, 572.0, 3432.0, -12012.0, 30030.0, 45045.0,  // C4[1], coeff of eps^5, polynomial in n of order 0
            1.0, 9009.0,  // C4[1], coeff of eps^4, polynomial in n of order 1
            -2944.0, 468.0, 135135.0,  // C4[1], coeff of eps^3, polynomial in n of order 2
            5792.0, 1040.0, -1287.0, 135135.0,  // C4[1], coeff of eps^2, polynomial in n of order 3
            5952.0, -11648.0, 9152.0, -2574.0, 135135.0,  // C4[1], coeff of eps^1, polynomial in n of order 4
            -64.0, -624.0, 4576.0, -6864.0, 3003.0, 135135.0,  // C4[2], coeff of eps^5, polynomial in n of order 0
            8.0, 10725.0,  // C4[2], coeff of eps^4, polynomial in n of order 1
            1856.0, -936.0, 225225.0,  // C4[2], coeff of eps^3, polynomial in n of order 2
            -8448.0, 4992.0, -1144.0, 225225.0,  // C4[2], coeff of eps^2, polynomial in n of order 3
            -1440.0, 4160.0, -4576.0, 1716.0, 225225.0,  // C4[3], coeff of eps^5, polynomial in n of order 0
            -136.0, 63063.0,  // C4[3], coeff of eps^4, polynomial in n of order 1
            1024.0, -208.0, 105105.0,  // C4[3], coeff of eps^3, polynomial in n of order 2
            3584.0, -3328.0, 1144.0, 315315.0,  // C4[4], coeff of eps^5, polynomial in n of order 0
            -128.0, 135135.0,  // C4[4], coeff of eps^4, polynomial in n of order 1
            -2560.0, 832.0, 405405.0,  // C4[5], coeff of eps^5, polynomial in n of order 0
            128.0, 99099.0
        )
        var o = 0
        var k = 0
        for (l in 0 until nC4_) {        // l is index of C4[l]
            for (j in nC4_ - 1 downTo l) { // coeff of eps^j
                val m = nC4_ - j - 1 // order of polynomial in n
                _C4x[k++] = GeoMath.polyval(m, coeff, o, _n) / coeff[o + m + 1]
                o += m + 2
            }
        }
    }

    companion object {

        /**
         * A global instantiation of Geodesic with the parameters for the WGS84
         * ellipsoid.
         */
        @JvmField
        val WGS84 = Geodesic(Constants.WGS84_a, Constants.WGS84_f)

        /**
         * The order of the expansions used by Geodesic.
         */
        protected const val GEOGRAPHICLIB_GEODESIC_ORDER = 6
        protected const val nA1_ = GEOGRAPHICLIB_GEODESIC_ORDER
        const val nC1_ = GEOGRAPHICLIB_GEODESIC_ORDER
        const val nC1p_ = GEOGRAPHICLIB_GEODESIC_ORDER
        protected const val nA2_ = GEOGRAPHICLIB_GEODESIC_ORDER
        const val nC2_ = GEOGRAPHICLIB_GEODESIC_ORDER
        protected const val nA3_ = GEOGRAPHICLIB_GEODESIC_ORDER
        protected const val nA3x_ = nA3_
        const val nC3_ = GEOGRAPHICLIB_GEODESIC_ORDER
        protected const val nC3x_ = nC3_ * (nC3_ - 1) / 2
        const val nC4_ = GEOGRAPHICLIB_GEODESIC_ORDER
        protected const val nC4x_ = nC4_ * (nC4_ + 1) / 2
        private const val maxit1_ = 20
        private const val maxit2_ = maxit1_ + GeoMath.digits + 10

        // Underflow guard.  We require
        //   tiny_ * epsilon() > 0
        //   tiny_ + epsilon() == epsilon()
        val tiny_ = Math.sqrt(Math.MIN_NORMAL)
        private val tol0_ = Math.ulp(1.0)

        // Increase multiplier in defn of tol1_ from 100 to 200 to fix inverse case
        // 52.784459512564 0 -52.784459512563990912 179.634407464943777557
        // which otherwise failed for Visual Studio 10 (Release and Debug)
        private val tol1_ = 200 * tol0_
        private val tol2_ = Math.sqrt(tol0_)

        // Check on bisection interval
        private val tolb_ = tol0_
        private val xthresh_ = 1000 * tol2_

        // This is a reformulation of the geodesic problem.  The notation is as
        // follows:
        // - at a general point (no suffix or 1 or 2 as suffix)
        //   - phi = latitude
        //   - beta = latitude on auxiliary sphere
        //   - omega = longitude on auxiliary sphere
        //   - lambda = longitude
        //   - alpha = azimuth of great circle
        //   - sigma = arc length along great circle
        //   - s = distance
        //   - tau = scaled distance (= sigma at multiples of pi/2)
        // - at northwards equator crossing
        //   - beta = phi = 0
        //   - omega = lambda = 0
        //   - alpha = alpha0
        //   - sigma = s = 0
        // - a 12 suffix means a difference, e.g., s12 = s2 - s1.
        // - s and c prefixes mean sin and cos
        fun SinCosSeries(
            sinp: Boolean,
            sinx: Double, cosx: Double,
            c: DoubleArray
        ): Double {
            // Evaluate
            // y = sinp ? sum(c[i] * sin( 2*i    * x), i, 1, n) :
            //            sum(c[i] * cos((2*i+1) * x), i, 0, n-1)
            // using Clenshaw summation.  N.B. c[0] is unused for sin series
            // Approx operation count = (n + 5) mult and (2 * n + 2) add
            var k = c.size
            // Point to one beyond last element
            var n = k - if (sinp) 1 else 0
            val ar = 2 * (cosx - sinx) * (cosx + sinx)
            // 2 * cos(2 * x)
            var y0 = if (n and 1 != 0) c[--k] else 0.0
            var y1 = 0.0 // accumulators for sum
            // Now n is even
            n /= 2
            while (n-- != 0) {
                // Unroll loop x 2, so accumulators return to their original role
                y1 = ar * y0 - y1 + c[--k]
                y0 = ar * y1 - y0 + c[--k]
            }
            return if (sinp) 2 * sinx * cosx * y0 // sin(2 * x) * y0
            else cosx * (y0 - y1) // cos(x) * (y0 - y1)
        }

        private fun Astroid(x: Double, y: Double): Double {
            // Solve k^4+2*k^3-(x^2+y^2-1)*k^2-2*y^2*k-y^2 = 0 for positive root k.
            // This solution is adapted from Geocentric::Reverse.
            val k: Double
            val p = GeoMath.sq(x)
            val q = GeoMath.sq(y)
            val r = (p + q - 1) / 6
            if (!(q == 0.0 && r <= 0)) {
                val   // Avoid possible division by zero when r = 0 by multiplying equations
                // for s and t by r^3 and r, resp.
                        S = p * q / 4
                // S = r^3 * s
                val r2 = GeoMath.sq(r)
                val r3 = r * r2
                // The discriminant of the quadratic equation for T3.  This is zero on
                // the evolute curve p^(1/3)+q^(1/3) = 1
                val disc = S * (S + 2 * r3)
                var u = r
                if (disc >= 0) {
                    var T3 = S + r3
                    // Pick the sign on the sqrt to maximize abs(T3).  This minimizes loss
                    // of precision due to cancellation.  The result is unchanged because
                    // of the way the T is used in definition of u.
                    T3 += if (T3 < 0) -Math.sqrt(disc) else Math.sqrt(disc) // T3 = (r * t)^3
                    // N.B. cbrt always returns the double root.  cbrt(-8) = -2.
                    val T = Math.cbrt(T3) // T = r * t
                    // T can be zero; but then r2 / T -> 0.
                    u += T + if (T != 0.0) r2 / T else 0.0
                } else {
                    // T is complex, but the way u is defined the result is double.
                    val ang = Math.atan2(Math.sqrt(-disc), -(S + r3))
                    // There are three possible cube roots.  We choose the root which
                    // avoids cancellation.  Note that disc < 0 implies that r < 0.
                    u += 2 * r * Math.cos(ang / 3)
                }
                val v = Math.sqrt(GeoMath.sq(u) + q)
                // guaranteed positive
                // Avoid loss of accuracy when u < 0.
                val uv = if (u < 0) q / (v - u) else u + v
                // u+v, guaranteed positive
                val w = (uv - q) / (2 * v) // positive?
                // Rearrange expression for k to avoid loss of accuracy due to
                // subtraction.  Division by 0 not possible because uv > 0, w >= 0.
                k = uv / (Math.sqrt(uv + GeoMath.sq(w)) + w) // guaranteed positive
            } else {               // q == 0 && r <= 0
                // y = 0 with |x| <= 1.  Handle this case directly.
                // for y small, positive root is k = abs(y)/sqrt(1-x^2)
                k = 0.0
            }
            return k
        }

        // The scale factor A1-1 = mean value of (d/dsigma)I1 - 1
        fun A1m1f(eps: Double): Double {
            val coeff = doubleArrayOf( // (1-eps)*A1-1, polynomial in eps2 of order 3
                1.0, 4.0, 64.0, 0.0, 256.0
            )
            val m = nA1_ / 2
            val t = GeoMath.polyval(m, coeff, 0, GeoMath.sq(eps)) / coeff[m + 1]
            return (t + eps) / (1 - eps)
        }

        // The coefficients C1[l] in the Fourier expansion of B1
        fun C1f(eps: Double, c: DoubleArray) {
            val coeff = doubleArrayOf( // C1[1]/eps^1, polynomial in eps2 of order 2
                -1.0, 6.0, -16.0, 32.0,  // C1[2]/eps^2, polynomial in eps2 of order 2
                -9.0, 64.0, -128.0, 2048.0,  // C1[3]/eps^3, polynomial in eps2 of order 1
                9.0, -16.0, 768.0,  // C1[4]/eps^4, polynomial in eps2 of order 1
                3.0, -5.0, 512.0,  // C1[5]/eps^5, polynomial in eps2 of order 0
                -7.0, 1280.0,  // C1[6]/eps^6, polynomial in eps2 of order 0
                -7.0, 2048.0
            )
            val eps2 = GeoMath.sq(eps)
            var d = eps
            var o = 0
            for (l in 1..nC1_) { // l is index of C1p[l]
                val m = (nC1_ - l) / 2 // order of polynomial in eps^2
                c[l] = d * GeoMath.polyval(m, coeff, o, eps2) / coeff[o + m + 1]
                o += m + 2
                d *= eps
            }
        }

        // The coefficients C1p[l] in the Fourier expansion of B1p
        fun C1pf(eps: Double, c: DoubleArray) {
            val coeff = doubleArrayOf( // C1p[1]/eps^1, polynomial in eps2 of order 2
                205.0, -432.0, 768.0, 1536.0,  // C1p[2]/eps^2, polynomial in eps2 of order 2
                4005.0, -4736.0, 3840.0, 12288.0,  // C1p[3]/eps^3, polynomial in eps2 of order 1
                -225.0, 116.0, 384.0,  // C1p[4]/eps^4, polynomial in eps2 of order 1
                -7173.0, 2695.0, 7680.0,  // C1p[5]/eps^5, polynomial in eps2 of order 0
                3467.0, 7680.0,  // C1p[6]/eps^6, polynomial in eps2 of order 0
                38081.0, 61440.0
            )
            val eps2 = GeoMath.sq(eps)
            var d = eps
            var o = 0
            for (l in 1..nC1p_) { // l is index of C1p[l]
                val m = (nC1p_ - l) / 2 // order of polynomial in eps^2
                c[l] = d * GeoMath.polyval(m, coeff, o, eps2) / coeff[o + m + 1]
                o += m + 2
                d *= eps
            }
        }

        // The scale factor A2-1 = mean value of (d/dsigma)I2 - 1
        fun A2m1f(eps: Double): Double {
            val coeff = doubleArrayOf( // (eps+1)*A2-1, polynomial in eps2 of order 3
                -11.0, -28.0, -192.0, 0.0, 256.0
            )
            val m = nA2_ / 2
            val t = GeoMath.polyval(m, coeff, 0, GeoMath.sq(eps)) / coeff[m + 1]
            return (t - eps) / (1 + eps)
        }

        // The coefficients C2[l] in the Fourier expansion of B2
        fun C2f(eps: Double, c: DoubleArray) {
            val coeff = doubleArrayOf( // C2[1]/eps^1, polynomial in eps2 of order 2
                1.0, 2.0, 16.0, 32.0,  // C2[2]/eps^2, polynomial in eps2 of order 2
                35.0, 64.0, 384.0, 2048.0,  // C2[3]/eps^3, polynomial in eps2 of order 1
                15.0, 80.0, 768.0,  // C2[4]/eps^4, polynomial in eps2 of order 1
                7.0, 35.0, 512.0,  // C2[5]/eps^5, polynomial in eps2 of order 0
                63.0, 1280.0,  // C2[6]/eps^6, polynomial in eps2 of order 0
                77.0, 2048.0
            )
            val eps2 = GeoMath.sq(eps)
            var d = eps
            var o = 0
            for (l in 1..nC2_) { // l is index of C2[l]
                val m = (nC2_ - l) / 2 // order of polynomial in eps^2
                c[l] = d * GeoMath.polyval(m, coeff, o, eps2) / coeff[o + m + 1]
                o += m + 2
                d *= eps
            }
        }
    }
}
